from elementals import hexDump
import socket
import struct
import time
import logging
from cryptography.hazmat.primitives.asymmetric.x25519   import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519  import Ed25519PrivateKey, Ed25519PublicKey
from cryptography.hazmat.primitives.ciphers             import Cipher, algorithms, modes
from cryptography.hazmat.primitives                     import serialization, hashes, hmac
from cryptography.hazmat.backends                       import default_backend
import base64
import secrets

import steamnetworkingsockets_messages_pb2
import steamnetworkingsockets_messages_udp_pb2
import steamnetworkingsockets_messages_certs_pb2

########################
# Protocol Definitions #
########################

UNENCRYPT_FORCE_ENCRYPT     = 0
UNENCRYPT_PREFER_ENCRYPT    = 1
UNENCRYPT_PREFER_CLEAR      = 2
UNENCRYPT_FORCE_CLEAR       = 3

UNENCRYPT_OPTIONS = {
                        UNENCRYPT_FORCE_ENCRYPT:  (steamnetworkingsockets_messages_pb2.k_ESteamNetworkingSocketsCipher_AES_256_GCM,),
                        UNENCRYPT_PREFER_ENCRYPT: (steamnetworkingsockets_messages_pb2.k_ESteamNetworkingSocketsCipher_AES_256_GCM, steamnetworkingsockets_messages_pb2.k_ESteamNetworkingSocketsCipher_NULL),
                        UNENCRYPT_PREFER_CLEAR:   (steamnetworkingsockets_messages_pb2.k_ESteamNetworkingSocketsCipher_NULL, steamnetworkingsockets_messages_pb2.k_ESteamNetworkingSocketsCipher_AES_256_GCM),
                        UNENCRYPT_FORCE_CLEAR:    (steamnetworkingsockets_messages_pb2.k_ESteamNetworkingSocketsCipher_NULL,),
                    }

UNENCRYPT_NAMES   = {
                        UNENCRYPT_FORCE_ENCRYPT:  "Force encryption",
                        UNENCRYPT_PREFER_ENCRYPT: "Prefer encryption",
                        UNENCRYPT_PREFER_CLEAR:   "Prefer cleartext",
                        UNENCRYPT_FORCE_CLEAR:    "Force cleartext",
                    }

KEY_TYPE_INVALID    = 0
KEY_TYPE_CURVE25519 = 1

OPCODE_CHALLENGE_REQUEST    = steamnetworkingsockets_messages_udp_pb2.k_ESteamNetworkingUDPMsg_ChallengeRequest  # Client  -> Server
OPCODE_CHALLENGE_REPLY      = steamnetworkingsockets_messages_udp_pb2.k_ESteamNetworkingUDPMsg_ChallengeReply    # Client <-  Server
OPCODE_CONNECTION_REQUEST   = steamnetworkingsockets_messages_udp_pb2.k_ESteamNetworkingUDPMsg_ConnectRequest    # Client  -> Server
OPCODE_CONNECTION_OK        = steamnetworkingsockets_messages_udp_pb2.k_ESteamNetworkingUDPMsg_ConnectOK         # Client <-  Server
OPCODE_CONNECTION_CLOSED    = steamnetworkingsockets_messages_udp_pb2.k_ESteamNetworkingUDPMsg_ConnectionClosed  # Client <-> Server
OPCODE_NO_CONNECTION        = steamnetworkingsockets_messages_udp_pb2.k_ESteamNetworkingUDPMsg_NoConnection      # Client <-> Server

OPCODE_NAMES                = {
                                OPCODE_CHALLENGE_REQUEST : "Challenge-Request",
                                OPCODE_CHALLENGE_REPLY:    "Challenge-Reply",
                                OPCODE_CONNECTION_REQUEST: "Connection-Request",
                                OPCODE_CONNECTION_OK:      "Connection-OK",
                                OPCODE_CONNECTION_CLOSED:  "Connection-Closed",
                                OPCODE_NO_CONNECTION:      "No-Connection",
                              }

PROTOCOL_PADDING_SIZE       = 0x200
PROTOCOL_PADDING_BYTE       = b"\x00"

# Latest supported version
PROTOCOL_VERSION            = 9

# same recv size as in the cpp code
RECV_MTU                    = 1300 + 1024

####################
# Config Variables #
####################

proto_connection_id = 0x1234
proto_ping_estimate = 0x22

cert_identity_string = "str:Exploit Client\0"
cert_app_ids         = (0, ) # Open source versions don't have an app_id API

# Use the same hardcoded "cert" key each time
cert_private_key     = Ed25519PrivateKey.from_private_bytes(bytes.fromhex("57C383AFF495BA0852BC9F01757DDA9E3CD8FAAAA5E6752D29F532DB0FA109B5"))
cert_public_key_raw  = cert_private_key.public_key().public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw)

# Generate a new exchange key for each connection
crypt_private_key    = X25519PrivateKey.generate()
crypt_public_key_raw = crypt_private_key.public_key().public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw)

####################
# Global Variables #
####################

logger              = None
server_address      = None

unencryption_type   = None

proto_msgs = {}

connection_id               = 0
connection_sequence_number  = 1

encryption_tx_key   = None
encryption_tx_iv    = None
encryption_rx_key   = None
encryption_rx_iv    = None

##
# Create a protocol message header
##
def protocolHeader(opcode, payload, padded):
    packet  = bytearray()
    packet += struct.pack("B", opcode)         # Also called "msg_id" in the code
    if padded:
        packet += struct.pack("<H", len(payload))
    packet += payload

    if padded and len(packet) < PROTOCOL_PADDING_SIZE:
        packet += PROTOCOL_PADDING_BYTE * (PROTOCOL_PADDING_SIZE - len(packet))
        logger.debug("Going to send the following (padded) packet:")
    else:
        logger.debug("Going to send the following packet:")
    logger.addIndent()
    logger.debug(hexDump(packet))
    logger.removeIndent()

    return packet

##
# Receives a protocol message
##
def recvProtocolMessage(expected_opcode, socket_fd):
    logger.info("Receiving a protocol message")

    packet, _ = socket_fd.recvfrom(RECV_MTU)
    logger.addIndent()
    logger.debug(hexDump(packet))
    logger.removeIndent()

    # minimal size checks
    if len(packet) < 1:
        logger.warning("Received a packet smaller than the minimal header size!")
        raise ValueError

    # parse the header
    opcode = struct.unpack("B", packet[:1])[0]
    if opcode != expected_opcode:
        logger.warning(f"Expected message {OPCODE_NAMES[expected_opcode]} while received {OPCODE_NAMES[opcode]}!")
        raise ValueError
    else:
        logger.info(f"Received the expected message: {OPCODE_NAMES[expected_opcode]}")
    return packet[1:]

##
# Construct our self ceritifcate
##
def buildCert():
    self_cert = steamnetworkingsockets_messages_certs_pb2.CMsgSteamDatagramCertificate()
    self_cert.identity_string = cert_identity_string
    self_cert.app_ids.extend(cert_app_ids)
    self_cert.key_type = KEY_TYPE_CURVE25519
    self_cert.key_data = cert_public_key_raw
    return self_cert

##
# Construct our signed ceritifcate
##
def buildSignedCert():
    self_cert = steamnetworkingsockets_messages_certs_pb2.CMsgSteamDatagramCertificateSigned()
    self_cert.cert = buildCert().SerializeToString()
    # The cert isn't really signed, we don't provide any signature for it

    return self_cert

##
# Construct our crypto info
##
def buildCryptoInfo():
    self_crypto = steamnetworkingsockets_messages_pb2.CMsgSteamDatagramSessionCryptInfo()
    self_crypto.key_type            = KEY_TYPE_CURVE25519
    self_crypto.key_data            = crypt_public_key_raw
    self_crypto.nonce               = secrets.randbelow(2 ** 64)
    self_crypto.protocol_version    = PROTOCOL_VERSION
    self_crypto.ciphers.extend(UNENCRYPT_OPTIONS[unencryption_type])

    return self_crypto

##
# Construct our signed crypto preferences
##
def buildSignedCryptoInfo():
    self_crypto = steamnetworkingsockets_messages_pb2.CMsgSteamDatagramSessionCryptInfoSigned();
    self_crypto.info        = buildCryptoInfo().SerializeToString()
    self_crypto.signature   = cert_private_key.sign(self_crypto.info)

    return self_crypto

##
# Sends the 1st protocol message - Challenge Request
##
def sendChallengeRequest(socket_fd):
    global proto_msgs

    opcode = OPCODE_CHALLENGE_REQUEST
    logger.info(f"Sending a {OPCODE_NAMES[opcode]} message")
    
    challenge_request = steamnetworkingsockets_messages_udp_pb2.CMsgSteamSockets_UDP_ChallengeRequest()
    challenge_request.connection_id     = proto_connection_id
    challenge_request.my_timestamp      = int(time.time() * 1000000) # Time in microseconds
    challenge_request.protocol_version  = PROTOCOL_VERSION
    logger.addIndent()
    logger.debug(f"{challenge_request}")
    logger.removeIndent()

    # store the message in the context
    proto_msgs[opcode] = challenge_request

    packet = protocolHeader(opcode, challenge_request.SerializeToString(), padded=True)
    # send the packet to the server
    socket_fd.sendto(packet, server_address)

##
# Receives the 2nd protocol message - Challenge Reply
##
def recvChallengeReply(socket_fd):
    global proto_msgs

    opcode = OPCODE_CHALLENGE_REPLY
    payload = recvProtocolMessage(opcode, socket_fd)
    
    challenge_reply = steamnetworkingsockets_messages_udp_pb2.CMsgSteamSockets_UDP_ChallengeReply()
    challenge_reply.ParseFromString(payload)
    logger.addIndent()
    logger.debug(f"{challenge_reply}")
    logger.removeIndent()

    # store the message in the context
    proto_msgs[opcode] = challenge_reply

##
# Sends the 3rd protocol message - Connection Request
##
def sendConnectionRequest(socket_fd):
    global proto_msgs

    opcode = OPCODE_CONNECTION_REQUEST
    logger.info(f"Sending a {OPCODE_NAMES[opcode]} message")

    prev_message = proto_msgs[OPCODE_CHALLENGE_REPLY]
    
    connection_request = steamnetworkingsockets_messages_udp_pb2.CMsgSteamSockets_UDP_ConnectRequest()
    connection_request.client_connection_id = proto_connection_id
    connection_request.challenge            = prev_message.challenge
    connection_request.ping_est_ms          = proto_ping_estimate
    connection_request.cert.CopyFrom(buildSignedCert())
    connection_request.crypt.CopyFrom(buildSignedCryptoInfo())
    logger.addIndent()
    logger.debug(f"{connection_request}")
    logger.removeIndent()

    # store the message in the context
    proto_msgs[opcode] = connection_request

    packet = protocolHeader(opcode, connection_request.SerializeToString(), padded=False)
    # send the packet to the server
    socket_fd.sendto(packet, server_address)

##
# Receives the 4th protocol message - Connection OK
##
def recvConnectionOK(socket_fd):
    global proto_msgs, connection_id

    opcode = OPCODE_CONNECTION_OK
    payload = recvProtocolMessage(opcode, socket_fd)
    
    connection_ok = steamnetworkingsockets_messages_udp_pb2.CMsgSteamSockets_UDP_ConnectOK()
    connection_ok.ParseFromString(payload)
    logger.addIndent()
    logger.debug(f"{connection_ok}")
    logger.removeIndent()

    # store the message in the context
    proto_msgs[opcode] = connection_ok

    # extract the connection_id
    connection_id = connection_ok.server_connection_id

##
# Derive the symmetric send/recv keys
##
def deriveCryptoKeys():
    global encryption_tx_key, encryption_rx_key, encryption_tx_iv, encryption_rx_iv

    # Collect all the needed information from the different messages
    self_crypt_raw  = proto_msgs[OPCODE_CONNECTION_REQUEST].crypt.info
    self_cert_raw   = proto_msgs[OPCODE_CONNECTION_REQUEST].cert.cert
    peer_crypt_raw  = proto_msgs[OPCODE_CONNECTION_OK].crypt.info
    peer_cert_raw   = proto_msgs[OPCODE_CONNECTION_OK].cert.cert

    self_crypt = steamnetworkingsockets_messages_pb2.CMsgSteamDatagramSessionCryptInfo()
    self_crypt.ParseFromString(self_crypt_raw)

    peer_crypt = steamnetworkingsockets_messages_pb2.CMsgSteamDatagramSessionCryptInfo()
    peer_crypt.ParseFromString(peer_crypt_raw)
    
    logger.debug("Generate the premaster shared secret (DH Exchange)")
    peer_public_key = X25519PublicKey.from_public_bytes(peer_crypt.key_data)
    prepremaster_key = crypt_private_key.exchange(peer_public_key)
    h = hashes.Hash(hashes.SHA256(), backend=default_backend())
    h.update(prepremaster_key)
    premaster_key = h.finalize()
    logger.debug(hexDump(premaster_key))

    logger.debug("Extract the shared secret and expand it using both nonce values")
    h = hmac.HMAC(premaster_key, hashes.SHA256(), backend=default_backend())
    h.update(struct.pack("<QQ", peer_crypt.nonce, self_crypt.nonce))
    shared_key = h.finalize()
    logger.debug(hexDump(shared_key))

    logger.debug("Derive the keys based on additional exchanged fields")
    connection_ids = struct.pack("<LL", proto_connection_id, connection_id)
    auxilery = connection_ids + b"Steam datagram" + peer_cert_raw + self_cert_raw + peer_crypt_raw + self_crypt_raw

    round_keys = []
    for idx in range(4):
        h = hmac.HMAC(shared_key, hashes.SHA256(), backend=default_backend())
        h.update(auxilery)
        h.update(struct.pack("B", idx + 1))
        round_key = h.finalize()
        auxilery = round_key + (auxilery if idx == 0 else auxilery[len(round_key):])
        round_keys.append(round_key)

    # now sort the keys that we created
    encryption_tx_key = round_keys[0]
    encryption_rx_key = round_keys[1]
    encryption_tx_iv  = round_keys[2][:12]
    encryption_rx_iv  = round_keys[3][:12]

    logger.debug("TX Key:")
    logger.debug(hexDump(encryption_tx_key))
    logger.debug("RX Key:")
    logger.debug(hexDump(encryption_rx_key))
    logger.debug("TX IV:")
    logger.debug(hexDump(encryption_tx_iv))
    logger.debug("RX IV:")
    logger.debug(hexDump(encryption_rx_iv))

##
# Create a Steam "Connection" with the target server
##
def createSteamConnection(server_ip, server_port, use_encryption, log):
    global logger, server_address, unencryption_type

    logger              = log
    server_address      = (server_ip, server_port)
    unencryption_type   = UNENCRYPT_FORCE_ENCRYPT if use_encryption else UNENCRYPT_FORCE_CLEAR

    logger.info("Creating a Steam connection with the server")
    logger.addIndent()
    logger.info(f"Server address is: {server_ip}:{server_port}")
    logger.info(f"Client unencryption support: {UNENCRYPT_NAMES[unencryption_type]}")

    # create the UDP socket to be used for the connection
    socket_fd = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    # send the 1st handshake packet to the server
    sendChallengeRequest(socket_fd)
    # receive the 2nd handshake packet back from the server
    recvChallengeReply(socket_fd)
    # send the 3rd handshake packet to the server
    sendConnectionRequest(socket_fd)
    # receive the 4th handshake packet back from the server
    recvConnectionOK(socket_fd)

    # derive encryption keyrs for send/recv
    if unencryption_type == UNENCRYPT_FORCE_ENCRYPT:
        logger.info("Derive the send/recv encryption keys")
        logger.addIndent()
        deriveCryptoKeys()
        logger.removeIndent()

    logger.info("Connection created successfully!")
    logger.removeIndent()

    # return back the required socket
    return socket_fd

##
# Create a connection message header
##
def connectionSend(socket_fd, payload, protobuf=b''):
    global connection_sequence_number

    packet  = bytearray()
    packet += struct.pack("B", 0x80 | (0 if len(protobuf) == 0 else 1))   # data message
    packet += struct.pack("<L", connection_id)                            # connection id
    packet += struct.pack("<H", connection_sequence_number & 0xFFFF)      # sequence number

    # the protobuf part is unencrypted
    packet += protobuf

    # encrypt the outgoing payload (encryption is ON)
    if unencryption_type == UNENCRYPT_FORCE_ENCRYPT:
        iv = struct.pack("<Q", struct.unpack("<Q", encryption_tx_iv[:8])[0] + connection_sequence_number) + encryption_tx_iv[8:]
        encryptor = Cipher(algorithms.AES(encryption_tx_key), modes.GCM(iv), backend=default_backend()).encryptor()
        packet += encryptor.update(payload) + encryptor.finalize() + encryptor.tag
    else:
        packet += payload

    # advance the seq number
    connection_sequence_number += 1

    # explicit if check due to performance issues
    if logger._min_level == logging.DEBUG:
        logger.debug("Going to send the following packet:")
        logger.addIndent()
        logger.debug(hexDump(packet))
        logger.removeIndent()

    socket_fd.sendto(packet, server_address)
